"""Root of the quiz API.

The entire public API is available at root level:

   >>> import quiz
   >>> quiz.Schema, quiz.execute, quiz.SelectionError, ...
"""
from .__about__ import (__author__, __copyright__, __description__,  # noqa
                        __version__)
from .build import *  # noqa
from .execution import *  # noqa
from .schema import *  # noqa
from .types import *  # noqa

from . import build, schema  # noqa isort:skip



import sys

PY3 = sys.version_info > (3, )
HAS_PEP519 = sys.version_info > (3, 6)
PY2 = not PY3


if PY3:
    from functools import singledispatch
    from textwrap import indent
    map = map
else:  # pragma: no cover
    from singledispatch import singledispatch  # noqa

    def indent(text, pad):
        return '\n'.join(map(pad.__add__, text.splitlines()))

    from itertools import imap as map


if HAS_PEP519:
    from os import fspath
else:  # pragma: no cover
    # code adapted from https://www.python.org/dev/peps/pep-0519/
    def fspath(path):
        # type: Union[Pathlike, str, bytes] -> Union[str, bytes]
        """Return the string representation of the path.

        If str or bytes is passed in, it is returned unchanged. If __fspath__()
        returns something other than str or bytes then TypeError is raised. If
        this function is given something that is not str, bytes, or os.PathLike
        then TypeError is raised.
        """
        if isinstance(path, (str, bytes)):
            return path

        # Work from the object's type to match method resolution of other magic
        # methods.
        path_type = type(path)
        try:
            path = path_type.__fspath__(path)
        except AttributeError:
            if hasattr(path_type, '__fspath__'):
                raise
        else:
            if isinstance(path, (str, bytes)):
                return path
            else:
                raise TypeError(
                    "expected __fspath__() to return str or bytes, "
                    "not " + type(path).__name__)

        raise TypeError("expected str, bytes or os.PathLike object, not "
                        + path_type.__name__)



"""Components for executing GraphQL operations"""
import json
import typing as t
from functools import partial

import snug
from gentools import irelay, py2_compatible, return_

from .build import Query
from .types import load
from .utils import JSON, ValueObject

__all__ = [
    'execute',
    'execute_async',
    'executor',
    'async_executor',

    'Executable',

    'ErrorResponse',
    'HTTPError',
]

Executable = t.Union[str, Query]
"""Anything which can be executed as a GraphQL operation"""


@py2_compatible
def _exec(executable):
    # type: Executable -> Generator
    if isinstance(executable, str):
        return_((yield executable))
    elif isinstance(executable, Query):
        return_(load(
            executable.cls,
            executable.selections,
            (yield str(executable))
        ))
    else:
        raise NotImplementedError('not executable: ' + repr(executable))


@py2_compatible
def middleware(url, query_str):
    # type: (str, str) -> snug.Query[Dict[str, JSON]]
    response = yield snug.Request(
        'POST',
        url,
        content=json.dumps({'query': query_str}).encode('ascii'),
        headers={'Content-Type': 'application/json'}
    )
    if response.status_code >= 400:
        raise HTTPError(response)
    content = json.loads(response.content.decode('utf-8'))
    if 'errors' in content:
        content.setdefault('data', {})
        raise ErrorResponse(**content)
    return_(content['data'])


def execute(obj, url, **kwargs):
    """Execute a GraphQL executable

    Parameters
    ----------
    obj: :data:`~quiz.execution.Executable`
        The object to execute.
        This may be a raw string or a query
    url: str
        The URL of the target endpoint
    **kwargs
         ``auth`` and/or ``client``, passed to :func:`snug.query.execute`.

    Returns
    -------
    JSON
        The response data

    Raises
    ------
    ErrorResponse
        If errors are present in the response
    HTTPError
        If the response has a non 2xx response code
    """
    snug_query = irelay(_exec(obj), partial(middleware, url))
    return snug.execute(snug_query, **kwargs)


def executor(**kwargs):
    """Create a version of :func:`execute` with bound arguments.
    Equivalent to ``partial(execute, **kwargs)``.

    Parameters
    ----------
    **kwargs
       ``url``, ``auth``, and/or ``client``, passed to :func:`execute`

    Returns
    -------
    ~typing.Callable[[Executable], JSON]
        A callable to execute GraphQL executables

    Example
    -------

    >>> execute = executor(url='https://api.github.com/graphql',
    ...                    auth=('me', 'password'))
    >>> result = execute('''
    ...   {
    ...     repository(owner: "octocat" name: "Hello-World") {
    ...       description
    ...     }
    ...   }
    ... ''', client=requests.Session())
    """
    return partial(execute, **kwargs)


def execute_async(obj, url, **kwargs):
    """Execute a GraphQL executable asynchronously

    Parameters
    ----------
    obj: Executable
        The object to execute.
        This may be a raw string or a query
    url: str
        The URL of the target endpoint
    **kwargs
         ``auth`` and/or ``client``,
         passed to :func:`snug.query.execute_async`.

    Returns
    -------
    JSON
        The response data

    Raises
    ------
    ErrorResponse
        If errors are present in the response
    HTTPError
        If the response has a non 2xx response code
    """
    snug_query = irelay(_exec(obj), partial(middleware, url))
    return snug.execute_async(snug_query, **kwargs)


def async_executor(**kwargs):
    """Create a version of :func:`execute_async` with bound arguments.
    Equivalent to ``partial(execute_async, **kwargs)``.

    Parameters
    ----------
    **kwargs
       ``url``, ``auth``, and/or ``client``, passed to :func:`execute_async`

    Returns
    -------
    ~typing.Callable[[Executable], ~typing.Awaitable[JSON]]
        A callable to asynchronously execute GraphQL executables

    Example
    -------

    >>> execute = async_executor(url='https://api.github.com/graphql',
    ...                          auth=('me', 'password'))
    >>> result = await execute('''
    ...   {
    ...     repository(owner: "octocat" name: "Hello-World") {
    ...       description
    ...     }
    ...   }
    ... ''')
    """
    return partial(execute_async, **kwargs)


class ErrorResponse(ValueObject, Exception):
    """A response containing errors"""

    __fields__ = [
        ('data', t.Dict[str, JSON], 'Data returned in the response'),
        ('errors', t.List[t.Dict[str, JSON]],
         'Errors returned in the response'),
    ]


class HTTPError(ValueObject, Exception):
    __fields__ = [
        ('response', snug.Response, 'The response object'),
    ]

    def __str__(self):
        return ('Response with status {0.status_code}, content: {0.content!r}'
                .format(self.response))



import os.path

from setuptools import find_packages, setup


def read(fname):
    path = os.path.join(os.path.dirname(__file__), fname)
    with open(path, 'r') as rfile:
        return rfile.read()


metadata = {}
exec(read('quiz/__about__.py'), metadata)


setup(
    name='quiz',
    version=metadata['__version__'],
    description=metadata['__description__'],
    license=metadata['__license__'],
    long_description=read('README.rst') + '\n\n' + read('HISTORY.rst'),
    url='https://github.com/ariebovenberg/quiz',

    author=metadata['__author__'],
    author_email='a.c.bovenberg@gmail.com',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    install_requires=[
        'enum34~=1.1.6; python_version<"3.4"',
        'gentools~=1.1.0',
        'six>=1.11,<1.13',
        'snug~=1.3.4',
        'typing>=3.6.6; python_version<"3.5"',
    ],
    keywords=['graphql', 'http', 'async'],
    python_requires='>=2.7',
    packages=find_packages(exclude=('tests', 'docs')),
)



"""Components for typed GraphQL interactions"""
import enum
import typing as t
from itertools import starmap

import six

from .build import InlineFragment
from .utils import FrozenDict, ValueObject

__all__ = [
    # types
    'Enum',
    'Union',
    'GenericScalar',
    'Scalar',
    'List',
    'Interface',
    'Object',
    'Nullable',
    'FieldDefinition',
    'InputValue',
    # TODO: mutation
    # TODO: subscription

    # validation
    'validate',
    'ValidationError',
    'SelectionError',
    'NoSuchField',
    'NoSuchArgument',
    'SelectionsNotSupported',
    'InvalidArgumentType',
    'MissingArgument',

    'NoValueForField',
    'load',
]


InputValue = t.NamedTuple('InputValue', [
    ('name', str),
    ('desc', str),
    ('type', type),
])

_PRIMITIVE_TYPES = (int, float, bool, six.text_type)


class HasFields(type):
    """metaclass for classes with GraphQL field definitions"""

    def __getitem__(self, selection_set):
        # type: SelectionSet -> InlineFragment
        return InlineFragment(self, validate(self, selection_set))


class Namespace(object):

    def __init__(__self__, **kwargs):
        __self__.__dict__.update(kwargs)

    def __eq__(self, other):
        if type(self) == type(other):
            return self.__dict__ == other.__dict__
        return NotImplemented

    if six.PY2:  # pragma: no cover
        def __ne__(self, other):
            equal = self.__eq__(other)
            return NotImplemented if equal is NotImplemented else not equal

    def __repr__(self):
        return '{}({})'.format(
            getattr(self.__class__, '__qualname__' if six.PY3 else '__name__'),
            ', '.join(starmap('{}={!r}'.format, self.__dict__.items()))
        )


@six.add_metaclass(HasFields)
class Object(Namespace):
    """a graphQL object"""


class InputObject(object):
    """not yet implemented"""


# separate class to distinguish graphql enums from normal Enums
class Enum(enum.Enum):
    pass


class Interface(HasFields):
    """metaclass for interfaces"""


class NoValueForField(AttributeError):
    """Indicates a value cannot be retrieved for the field"""


class FieldDefinition(ValueObject):
    __fields__ = [
        ('name', str, 'Field name'),
        ('desc', str, 'Field description'),
        ('type', type, 'Field data type'),
        ('args', FrozenDict[str, InputValue], 'Accepted field arguments'),
        ('is_deprecated', bool, 'Whether the field is deprecated'),
        ('deprecation_reason', t.Optional[str], 'Reason for deprecation'),
    ]

    def __get__(self, obj, objtype=None):
        if obj is None:  # accessing on class
            return self
        try:
            return obj.__dict__[self.name]
        except KeyError:
            raise NoValueForField()

    # full descriptor interface is necessary to be displayed nicely in help()
    def __set__(self, obj, value):
        raise AttributeError("Can't set field value")

    # __doc__ allows descriptor to be displayed nicely in help()
    @property
    def __doc__(self):
        return ': {.__name__}\n    {}'.format(self.type, self.desc)


class ListMeta(type):

    def __getitem__(self, arg):
        return type('[{.__name__}]'.format(arg), (List, ), {
            '__arg__': arg
        })

    def __instancecheck__(self, instance):
        return (isinstance(instance, list)
                and all(isinstance(i, self.__arg__) for i in instance))


# Q: why not typing.List?
# A: it doesn't support __doc__, __name__, or isinstance()
@six.add_metaclass(ListMeta)
class List(object):
    __arg__ = object


class NullableMeta(type):

    def __getitem__(self, arg):
        return type('{.__name__} or None'.format(arg), (Nullable, ), {
            '__arg__': arg
        })

    def __instancecheck__(self, instance):
        return instance is None or isinstance(instance, self.__arg__)


# Q: why not typing.Optional?
# A: it is not easily distinguished from Union,
#    and doesn't support __doc__, __name__, or isinstance()
@six.add_metaclass(NullableMeta)
class Nullable(object):
    __arg__ = object


class UnionMeta(type):

    def __instancecheck__(self, instance):
        return isinstance(instance, self.__args__)


# Q: why not typing.Union?
# A: it isn't consistent across python versions,
#    and doesn't support __doc__, __name__, or isinstance()
@six.add_metaclass(UnionMeta)
class Union(object):
    __args__ = ()


class Scalar(object):
    """Base class for scalars"""

    def __gql_dump__(self):
        """Serialize the scalar to a GraphQL primitive value"""
        raise NotImplementedError(
            'GraphQL serialization is not defined for this scalar')

    @classmethod
    def __gql_load__(cls, data):
        """Load a scalar instance from GraphQL"""
        raise NotImplementedError(
            'GraphQL deserialization is not defined for this scalar')


class GenericScalarMeta(type):

    def __instancecheck__(self, instance):
        return isinstance(instance, _PRIMITIVE_TYPES)


@six.add_metaclass(GenericScalarMeta)
class GenericScalar(Scalar):
    """A generic scalar, accepting any primitive type"""


def _unwrap_list_or_nullable(type_):
    # type: Type[Nullable, List, Scalar, Enum, InputObject]
    # -> Type[Scalar | Enum | InputObject]
    if issubclass(type_, (Nullable, List)):
        return _unwrap_list_or_nullable(type_.__arg__)
    return type_


def _validate_args(schema, actual):
    # type: (Mapping[str, InputValue], Mapping[str, object])
    # -> Mapping[str, object]
    invalid_args = six.viewkeys(actual) - six.viewkeys(schema)
    if invalid_args:
        raise NoSuchArgument(invalid_args.pop())

    for input_value in schema.values():
        try:
            value = actual[input_value.name]
        except KeyError:
            if issubclass(input_value.type, Nullable):
                continue  # arguments of nullable type may be omitted
            else:
                raise MissingArgument(input_value.name)

        if not isinstance(value, input_value.type):
            raise InvalidArgumentType(input_value.name, value)

    return actual


def _validate_field(schema, actual):
    # type (Optional[FieldDefinition], Field) -> Field
    # raises:
    # - NoSuchField
    # - SelectionsNotSupported
    # - NoSuchArgument
    # - RequredArgument
    # - InvalidArgumentType
    if schema is None:
        raise NoSuchField()
    _validate_args(schema.args, actual.kwargs)
    if actual.selection_set:
        type_ = _unwrap_list_or_nullable(schema.type)
        if not isinstance(type_, HasFields):
            raise SelectionsNotSupported()
        validate(type_, actual.selection_set)
    return actual


def validate(cls, selection_set):
    """Validate a selection set against a type

    Parameters
    ----------
    cls: type
        The class to validate against, an ``Object`` or ``Interface``
    selection_set: SelectionSet
        The selection set to validate

    Returns
    -------
    SelectionSet
        The validated selection set

    Raises
    ------
    SelectionError
        If the selection set is not valid
    """
    for field in selection_set:
        try:
            _validate_field(getattr(cls, field.name, None), field)
        except ValidationError as e:
            raise SelectionError(cls, field.name, e)
    return selection_set


# TODO: refactor using singledispatch
# TODO: cleanup this API: ``field`` is often unneeded. unify with ``load``?
def load_field(type_, field, value):
    # type: (Type[T], Field, JSON) -> T
    if issubclass(type_, Namespace):
        assert isinstance(value, dict)
        return load(type_, field.selection_set, value)
    elif issubclass(type_, Nullable):
        return None if value is None else load_field(
            type_.__arg__, field, value)
    elif issubclass(type_, List):
        assert isinstance(value, list)
        return [load_field(type_.__arg__, field, v) for v in value]
    elif issubclass(type_, _PRIMITIVE_TYPES):
        assert isinstance(value, type_)
        return value
    elif issubclass(type_, GenericScalar):
        assert isinstance(value, type_)
        return value
    elif issubclass(type_, Scalar):
        return type_.__gql_load__(value)
    elif issubclass(type_, Enum):
        assert value, type_._members_names_
        return type_(value)
    else:
        raise NotImplementedError()


def load(cls, selection_set, response):
    """Load a response for a selection set

    Parameters
    ----------
    cls: Type[T]
        The class to load against, an ``Object`` or ``Interface``
    selection_set: SelectionSet
        The selection set to validate
    response: t.Dict[str, JSON]
        The JSON response data

    Returns
    -------
    T
        An instance of ``cls``
    """
    return cls(**{
        field.alias or field.name: load_field(
            getattr(cls, field.name).type,
            field,
            response[field.alias or field.name],
        )
        for field in selection_set
    })


class ValidationError(Exception):
    """base class for validation errors"""


class SelectionError(ValueObject, ValidationError):
    __fields__ = [
        ('on', type, 'Type on which the error occurred'),
        ('path', str, 'Path at which the error occurred'),
        ('error', ValidationError, 'Original error'),
    ]

    def __str__(self):
        return '{} on "{}" at path "{}":\n\n    {}: {}'.format(
            self.__class__.__name__,
            self.on.__name__,
            self.path,
            self.error.__class__.__name__,
            self.error,
        )


class NoSuchField(ValueObject, ValidationError):
    __fields__ = []

    def __str__(self):
        return 'field does not exist'


class NoSuchArgument(ValueObject, ValidationError):
    __fields__ = [
        ('name', str, '(Invalid) argument name'),
    ]

    def __str__(self):
        return 'argument "{}" does not exist'.format(self.name)


class InvalidArgumentType(ValueObject, ValidationError):
    __fields__ = [
        ('name', str, 'Argument name'),
        ('value', object, '(Invalid) value'),
    ]

    def __str__(self):
        return 'invalid value "{}" of type {} for argument "{}"'.format(
            self.value,
            type(self.value),
            self.name,
        )


class MissingArgument(ValueObject, ValidationError):
    __fields__ = [
        ('name', str, 'Missing argument name'),
    ]

    def __str__(self):
        return 'argument "{}" missing (required)'.format(self.name)


class SelectionsNotSupported(ValueObject, ValidationError):
    __fields__ = []

    def __str__(self):
        return 'selections not supported on this object'


BUILTIN_SCALARS = {
    "Boolean": bool,
    "String":  str,
    "Float":   float,
    "Int":     int,
}



"""Main module for constructing graphQL queries"""
import enum
import re
import typing as t
from operator import attrgetter, methodcaller

from .compat import indent, singledispatch
from .utils import FrozenDict, ValueObject, compose, init_last

__all__ = [
    # building graphQL documents
    'SelectionSet',
    'Selection',
    'Field',
    'InlineFragment',
    'Raw',
    'Query',
    'SELECTOR',

    # render
    'gql',
    'escape',
    'argument_as_gql',
]

INDENT = "  "

gql = methodcaller("__gql__")


class SelectionSet(t.Iterable['Selection'], t.Sized):
    """Sequence of selections

    Parameters
    ----------
    *selections: Selection
        Items in the selection set.

    Notes
    -----
    * Instances are immutable.
    * Extending selection sets is possible through special methods
      (``__getattr__``, ``__call__``, ``__getitem__``)
    """
    # The attribute needs to have a dunder name to prevent
    # conflicts with GraphQL field names.
    # This is also why we can't just subclass `tuple`.
    __slots__ = '__selections__'

    def __init__(self, *selections):
        self.__selections__ = selections

    # TODO: check if actually faster
    # faster, internal, alternative to __init__
    @classmethod
    def _make(cls, selections):
        instance = cls.__new__(cls)
        instance.__selections__ = tuple(selections)
        return instance

    def __getattr__(self, fieldname):
        """Add a new field to the selection set.

        Parameters
        ----------
        fieldname: str
            The name of the field to add.

        Returns
        -------
        SelectionSet
            A selection set with the new field added to the end.

        Example
        -------

        This functionality can be used to quickly create a sequence of fields:

        >>> _ = SelectionSet()
        >>> str(
        ...     _
        ...     .foo
        ...     .bar
        ...     .bing
        ... )
        {
          foo
          bar
          bing
        }
        """
        return SelectionSet._make(self.__selections__ + (Field(fieldname), ))

    def __getitem__(self, selections):
        """Add a sub-selection to the last field in the selection set

        Parameters
        ----------
        selections: SelectionSet
            The selection set to nest

        Example
        -------

        >>> _ = SelectionSet()
        >>> str(
        ...     _
        ...     .foo
        ...     .bar[
        ...         _
        ...         .qux
        ...         .bing
        ...     ]
        ...     .other_field
        ... )
        {
          foo
          bar {
            qux
            bing
          }
          other_field
        }

        Returns
        -------
        SelectionSet
            A selection set with selections added to the last field.

        Raises
        ------
        utils.Empty
            In case the selection set is empty
        """
        rest, target = init_last(self.__selections__)

        assert isinstance(selections, SelectionSet)
        assert len(selections.__selections__) >= 1

        return SelectionSet._make(
            tuple(rest)
            + (target.replace(selection_set=selections), ))

    def __repr__(self):
        return "<SelectionSet> {}".format(gql(self))

    # Positional arguments are retrieved manually from *args.
    # This ensures there can be no conflict with (named) **kwargs.
    # Otherwise, something like `self` could not be given as a keyword arg.
    def __call__(*args, **kwargs):
        """The selection set may be called in two distinct ways:

        1. With keyword arguments ``**kwargs``.
           These will be added as arguments to the last field
           in the selection set.
        2. With a single ``alias`` argument. This has the affect of adding
           an alias to the next field in the selection set.

        Parameters
        ----------
        alias: str, optional
            If given, the next field in the selection set will get this alias.

            Example
            -------

            >>> _ = SelectionSet()
            >>> str(
            ...     _
            ...     .foo
            ...     ('my_alias').bla
            ...     .other_field
            ... )
            {
               foo
               my_alias: bla
               other_field
            }

            Note
            ----
            The alias can only be specified as a :term:`positional argument`,
            and may not be combined with ``**kwargs``.

        **kwargs
            Adds arguments to the previous field in the chain

            Example
            -------

            >>> _ = SelectionSet()
            >>> str(
            ...     _
            ...     .foo
            ...     .bla(a=4, b='qux')
            ...     .other_field
            ... )
            {
              foo
              bla(a: 4, b: "qux")
              other_field
            }

            Note
            ----
            Each field argument must be a :term:`keyword argument`.

        Returns
        -------
        SelectionSet
            The new selection set

        Raises
        ------
        utils.Empty
            In case field arguments are given, but the selection set is empty
        """
        # TODO: check alias validity
        try:
            self, alias = args
        except ValueError:
            # alias is *not* given --> case 1
            self, = args
            return self.__add_kwargs(kwargs)
        else:
            # alias is given --> case 2
            return _AliasForNextField(*args)

    def __add_kwargs(self, args):
        rest, target = init_last(self.__selections__)
        return SelectionSet._make(
            tuple(rest) + (target.replace(kwargs=FrozenDict(args)), ))

    def __iter__(self):
        """Iterate over the selection set contents

        Returns
        -------
        Iterator[Selection]
            An iterator over selections
        """
        return iter(self.__selections__)

    def __len__(self):
        """Number of items in the selection set

        Returns
        -------
        int
            The number of items in the selection set
        """
        return len(self.__selections__)

    def __str__(self):
        """The selection set as raw graphQL"""
        return self.__gql__()

    def __gql__(self):
        return '{{\n{}\n}}'.format(
            '\n'.join(indent(gql(f), INDENT) for f in self)
        ) if self.__selections__ else ''

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return other.__selections__ == self.__selections__
        return NotImplemented

    def __ne__(self, other):
        equality = self.__eq__(other)
        return NotImplemented if equality is NotImplemented else not equality

    __hash__ = property(attrgetter('__selections__.__hash__'))


class _AliasForNextField(object):
    __slots__ = '__selection_set', '__alias'

    def __init__(self, selection_set, alias):
        self.__selection_set = selection_set
        self.__alias = alias

    def __getattr__(self, fieldname):
        return SelectionSet._make(
            self.__selection_set.__selections__
            + (Field(fieldname, alias=self.__alias), )
        )


SELECTOR = SelectionSet()
"""An empty, extendable :class:`SelectionSet`"""


class Raw(ValueObject):
    __fields__ = [
        ('content', str, 'The raw GraphQL content')
    ]

    def __gql__(self):
        return self.content


class Field(ValueObject):
    __fields__ = [
        ('name', str, 'Field name'),
        ('kwargs', FrozenDict, 'Given arguments'),
        ('selection_set', SelectionSet, 'Selection of subfields'),
        ('alias', t.Optional[str], 'Field alias'),
        # in the future:
        # - directives
    ]
    __defaults__ = (FrozenDict.EMPTY, SelectionSet(), None)

    def __gql__(self):
        arguments = '({})'.format(
            ', '.join(
                "{}: {}".format(k, argument_as_gql(v))
                for k, v in self.kwargs.items()
            )
        ) if self.kwargs else ''
        selection_set = (
            ' ' + gql(self.selection_set)
            if self.selection_set else '')
        alias = self.alias + ': ' if self.alias else ''
        return alias + self.name + arguments + selection_set


class InlineFragment(ValueObject):
    __fields__ = [
        ('on', type, 'Type of the fragment'),
        ('selection_set', SelectionSet, 'Subfields of the fragment'),
    ]
    # in the future: directives

    def __gql__(self):
        return '... on {} {}'.format(
            self.on.__name__,
            gql(self.selection_set)
        )


class Query(ValueObject):
    __fields__ = [
        ('cls', type, 'The query class'),
        ('selections', SelectionSet, 'Fields selection')
    ]
    # in the future:
    # - name (optional)
    # - variable_defs (optional)
    # - directives (optional)

    def __gql__(self):
        return 'query ' + gql(self.selections)

    def __str__(self):
        return self.__gql__()


_ESCAPE_PATTERNS = {
    '\b': r'\b',
    '\f': r'\f',
    '\n': r'\n',
    '\r': r'\r',
    '\t': r'\t',
    '\\': r'\\',
    '"':  r'\"',
}
_ESCAPE_RE = re.compile('|'.join(map(re.escape, _ESCAPE_PATTERNS)))


def _escape_match(match):
    return _ESCAPE_PATTERNS[match.group(0)]


def escape(txt):
    """Escape a string according to GraphQL specification

    Parameters
    ----------
    txt: str
        The string to escape

    Returns
    -------
    str
        the escaped string
    """
    return _ESCAPE_RE.sub(_escape_match, txt)


@singledispatch
def argument_as_gql(obj):
    # type: object -> str
    try:
        # consistent with other dunder methods, we look it up on the class
        serializer = type(obj).__gql_dump__
    except AttributeError:
        raise TypeError("Cannot serialize to GraphQL: {}".format(type(obj)))
    else:
        return serializer(obj)


argument_as_gql.register(str, compose('"{}"'.format, escape))
argument_as_gql.register(int, str)
argument_as_gql.register(type(None), 'null'.format)
argument_as_gql.register(bool, {True: 'true', False: 'false'}.__getitem__)
argument_as_gql.register(float, str)


@argument_as_gql.register(enum.Enum)
def _enum_to_gql(obj):
    return obj.value


Selection = t.Union[Field, InlineFragment]
"""Field or inline fragment"""
